;;; gnus-load.el --- automatically extracted custom dependencies
;;
;;; Code:

(put 'nnmail 'custom-loads '("nnmail"))
(put 'gnus-article-emphasis 'custom-loads '("gnus-art"))
(put 'gnus-article-headers 'custom-loads '("gnus-sum" "gnus-art"))
(put 'nnmail-procmail 'custom-loads '("nnmail"))
(put 'gnus-score-kill 'custom-loads '("gnus-kill"))
(put 'gnus-visual 'custom-loads '("smiley" "gnus" "gnus-picon" "gnus-art" "earcon"))
(put 'gnus-score-expire 'custom-loads '("gnus-score" "gnus-kill"))
(put 'gnus-summary-maneuvering 'custom-loads '("gnus-sum"))
(put 'gnus-start 'custom-loads '("gnus" "gnus-util" "gnus-start" "gnus-int" "gnus-group"))
(put 'gnus-extract-view 'custom-loads '("gnus-uu" "gnus-sum"))
(put 'gnus-various 'custom-loads '("gnus-sum"))
(put 'gnus-article-washing 'custom-loads '("gnus-art"))
(put 'gnus-score-files 'custom-loads '("gnus-score"))
(put 'message-news 'custom-loads '("message"))
(put 'gnus-thread 'custom-loads '("gnus-sum"))
(put 'languages 'custom-loads '("cus-edit"))
(put 'development 'custom-loads '("cus-edit"))
(put 'nnmail-various 'custom-loads '("nnmail"))
(put 'extensions 'custom-loads '("wid-edit"))
(put 'message-various 'custom-loads '("message"))
(put 'gnus-summary-exit 'custom-loads '("gnus-sum"))
(put 'news 'custom-loads '("message" "gnus"))
(put 'gnus 'custom-loads '("nnmail" "gnus" "gnus-win" "gnus-uu" "gnus-eform" "gnus-dup" "gnus-demon" "gnus-cache" "gnus-async" "gnus-art"))
(put 'gnus-summary-visual 'custom-loads '("gnus-sum"))
(put 'gnus-group-listing 'custom-loads '("gnus-group"))
(put 'gnus-score 'custom-loads '("gnus" "gnus-nocem"))
(put 'gnus-group-select 'custom-loads '("gnus-sum"))
(put 'message-buffers 'custom-loads '("message"))
(put 'gnus-threading 'custom-loads '("gnus-sum"))
(put 'gnus-score-decay 'custom-loads '("gnus-score"))
(put 'help 'custom-loads '("cus-edit"))
(put 'gnus-nocem 'custom-loads '("gnus-nocem"))
(put 'gnus-cite 'custom-loads '("gnus-cite"))
(put 'gnus-demon 'custom-loads '("gnus-demon"))
(put 'gnus-message 'custom-loads '("message"))
(put 'gnus-score-delta-default 'custom-loads '("gnus-sum" "gnus-score"))
(put 'nnmail-duplicate 'custom-loads '("nnmail"))
(put 'message-interface 'custom-loads '("message"))
(put 'nnmail-files 'custom-loads '("nnmail"))
(put 'gnus-edit-form 'custom-loads '("gnus-eform"))
(put 'emacs 'custom-loads '("cus-edit"))
(put 'gnus-summary-mail 'custom-loads '("gnus-sum"))
(put 'gnus-topic 'custom-loads '("gnus-topic"))
(put 'wp 'custom-loads '("cus-edit"))
(put 'gnus-summary-choose 'custom-loads '("gnus-sum"))
(put 'widget-browse 'custom-loads '("wid-browse"))
(put 'external 'custom-loads '("cus-edit"))
(put 'message-headers 'custom-loads '("message"))
(put 'message-forwarding 'custom-loads '("message"))
(put 'message-faces 'custom-loads '("message"))
(put 'environment 'custom-loads '("cus-edit"))
(put 'gnus-article-mime 'custom-loads '("gnus-sum" "gnus-art"))
(put 'gnus-duplicate 'custom-loads '("gnus-dup"))
(put 'nnmail-retrieve 'custom-loads '("nnmail"))
(put 'widgets 'custom-loads '("wid-edit" "wid-browse"))
(put 'earcon 'custom-loads '("earcon"))
(put 'hypermedia 'custom-loads '("wid-edit"))
(put 'gnus-group-levels 'custom-loads '("gnus-group"))
(put 'gnus-summary-format 'custom-loads '("gnus-sum"))
(put 'gnus-files 'custom-loads '("nnmail" "gnus"))
(put 'gnus-windows 'custom-loads '("gnus-win"))
(put 'gnus-article-buttons 'custom-loads '("gnus-art"))
(put 'gnus-summary 'custom-loads '("gnus" "gnus-sum"))
(put 'gnus-article-hiding 'custom-loads '("gnus-sum" "gnus-art"))
(put 'gnus-group 'custom-loads '("gnus" "gnus-topic"))
(put 'gnus-article-various 'custom-loads '("gnus-sum" "gnus-art"))
(put 'gnus-summary-marks 'custom-loads '("gnus-sum"))
(put 'gnus-article-saving 'custom-loads '("gnus-art"))
(put 'nnmail-expire 'custom-loads '("nnmail"))
(put 'message-mail 'custom-loads '("message"))
(put 'faces 'custom-loads '("wid-edit" "cus-edit" "message" "gnus"))
(put 'gnus-summary-various 'custom-loads '("gnus-sum"))
(put 'applications 'custom-loads '("cus-edit"))
(put 'gnus-extract-archive 'custom-loads '("gnus-uu"))
(put 'message 'custom-loads '("message"))
(put 'message-sending 'custom-loads '("message"))
(put 'editing 'custom-loads '("cus-edit"))
(put 'gnus-score-adapt 'custom-loads '("gnus-score"))
(put 'message-insertion 'custom-loads '("message"))
(put 'gnus-extract-post 'custom-loads '("gnus-uu"))
(put 'mail 'custom-loads '("message" "gnus"))
(put 'gnus-summary-sort 'custom-loads '("gnus-sum"))
(put 'customize 'custom-loads '("wid-edit" "custom" "cus-face" "cus-edit"))
(put 'nnmail-split 'custom-loads '("nnmail"))
(put 'gnus-asynchronous 'custom-loads '("gnus-async"))
(put 'gnus-article-highlight 'custom-loads '("gnus-art"))
(put 'gnus-extract 'custom-loads '("gnus-uu"))
(put 'gnus-article 'custom-loads '("gnus-cite" "gnus-art"))
(put 'gnus-group-foreign 'custom-loads '("gnus-group"))
(put 'programming 'custom-loads '("cus-edit"))
(put 'nnmail-prepare 'custom-loads '("nnmail"))
(put 'picons 'custom-loads '("gnus-picon"))
(put 'gnus-article-signature 'custom-loads '("gnus-art"))
(put 'gnus-group-various 'custom-loads '("gnus-group"))

(provide 'gnus-load)

;;; gnus-load.el ends here
